@@ -1,4 +1,4 @@
-export PYTEST_OPTIONS="--verbose -r s --timeout-method=thread --timeout=300 --durations=20"
+export PYTEST_OPTIONS="-r s --timeout-method=thread --timeout=300 --durations=20"
 if [[ $RUNSLOW != false ]]; then
     export PYTEST_OPTIONS="$PYTEST_OPTIONS --runslow"
 fi
@@ -19,7 +19,7 @@ echo "--"
 ulimit -a -H
 
 if [[ $COVERAGE == true ]]; then
-    coverage run $(which py.test) distributed -m "not avoid_travis" $PYTEST_OPTIONS;
+    coverage run $(which py.test) distributed -m "not avoid_travis" $PYTEST_OPTIONS
 else
-    py.test -m "not avoid_travis" distributed $PYTEST_OPTIONS;
-fi;
+    py.test -m "not avoid_travis" distributed $PYTEST_OPTIONS
+fi
