@@ -1,10 +1,11 @@
-from __future__ import print_function, division, absolute_import
-
 import errno
 import logging
 import socket
+from ssl import SSLError
 import struct
 import sys
+from tornado import gen
+import weakref
 
 try:
     import ssl
@@ -12,48 +13,38 @@ except ImportError:
     ssl = None
 
 import dask
-import tornado
-from tornado import gen, netutil
-from tornado.iostream import StreamClosedError, IOStream
+from tornado import netutil
+from tornado.iostream import StreamClosedError
 from tornado.tcpclient import TCPClient
 from tornado.tcpserver import TCPServer
 
-from ..compatibility import finalize, PY3
-from ..utils import (ensure_bytes, ensure_ip, get_ip, get_ipv6, nbytes,
-                     parse_timedelta, shutting_down)
+from ..system import MEMORY_LIMIT
+from ..threadpoolexecutor import ThreadPoolExecutor
+from ..utils import ensure_ip, get_ip, get_ipv6, nbytes, parse_timedelta, shutting_down
 
 from .registry import Backend, backends
 from .addressing import parse_host_port, unparse_host_port
 from .core import Comm, Connector, Listener, CommClosedError, FatalCommClosedError
-from .utils import (to_frames, from_frames,
-                    get_tcp_server_address, ensure_concrete_host)
+from .utils import to_frames, from_frames, get_tcp_server_address, ensure_concrete_host
 
 
 logger = logging.getLogger(__name__)
 
 
-def get_total_physical_memory():
-    try:
-        import psutil
-        return psutil.virtual_memory().total / 2
-    except ImportError:
-        return 2e9
+MAX_BUFFER_SIZE = MEMORY_LIMIT / 2
 
 
-MAX_BUFFER_SIZE = get_total_physical_memory()
-
-
-def set_tcp_timeout(stream):
+def set_tcp_timeout(comm):
     """
     Set kernel-level TCP timeout on the stream.
     """
-    if stream.closed():
+    if comm.closed():
         return
 
-    timeout = dask.config.get('distributed.comm.timeouts.tcp')
-    timeout = int(parse_timedelta(timeout, default='seconds'))
+    timeout = dask.config.get("distributed.comm.timeouts.tcp")
+    timeout = int(parse_timedelta(timeout, default="seconds"))
 
-    sock = stream.socket
+    sock = comm.socket
 
     # Default (unsettable) value on Windows
     # https://msdn.microsoft.com/en-us/library/windows/desktop/dd877220(v=vs.85).aspx
@@ -67,8 +58,7 @@ def set_tcp_timeout(stream):
 
     try:
         if sys.platform.startswith("win"):
-            logger.debug("Setting TCP keepalive: idle=%d, interval=%d",
-                         idle, interval)
+            logger.debug("Setting TCP keepalive: idle=%d, interval=%d", idle, interval)
             sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, idle * 1000, interval * 1000))
         else:
             sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
@@ -85,30 +75,33 @@ def set_tcp_timeout(stream):
                     TCP_KEEPIDLE = None
 
             if TCP_KEEPIDLE is not None:
-                logger.debug("Setting TCP keepalive: nprobes=%d, idle=%d, interval=%d",
-                             nprobes, idle, interval)
+                logger.debug(
+                    "Setting TCP keepalive: nprobes=%d, idle=%d, interval=%d",
+                    nprobes,
+                    idle,
+                    interval,
+                )
                 sock.setsockopt(socket.SOL_TCP, TCP_KEEPCNT, nprobes)
                 sock.setsockopt(socket.SOL_TCP, TCP_KEEPIDLE, idle)
                 sock.setsockopt(socket.SOL_TCP, TCP_KEEPINTVL, interval)
 
         if sys.platform.startswith("linux"):
-            logger.debug("Setting TCP user timeout: %d ms",
-                         timeout * 1000)
+            logger.debug("Setting TCP user timeout: %d ms", timeout * 1000)
             TCP_USER_TIMEOUT = 18  # since Linux 2.6.37
             sock.setsockopt(socket.SOL_TCP, TCP_USER_TIMEOUT, timeout * 1000)
     except EnvironmentError as e:
         logger.warning("Could not set timeout on TCP stream: %s", e)
 
 
-def get_stream_address(stream):
+def get_stream_address(comm):
     """
     Get a stream's local address.
     """
-    if stream.closed():
+    if comm.closed():
         return "<closed>"
 
     try:
-        return unparse_host_port(*stream.socket.getsockname()[:2])
+        return unparse_host_port(*comm.socket.getsockname()[:2])
     except EnvironmentError:
         # Probably EBADF
         return "<closed>"
@@ -122,33 +115,48 @@ def convert_stream_closed_error(obj, exc):
         # The stream was closed because of an underlying OS error
         exc = exc.real_error
         if ssl and isinstance(exc, ssl.SSLError):
-            if 'UNKNOWN_CA' in exc.reason:
-                raise FatalCommClosedError("in %s: %s: %s" % (obj, exc.__class__.__name__, exc))
-        raise CommClosedError("in %s: %s: %s" % (obj, exc.__class__.__name__, exc))
+            if "UNKNOWN_CA" in exc.reason:
+                raise FatalCommClosedError(
+                    "in %s: %s: %s" % (obj, exc.__class__.__name__, exc)
+                )
+        raise CommClosedError(
+            "in %s: %s: %s" % (obj, exc.__class__.__name__, exc)
+        ) from exc
     else:
-        raise CommClosedError("in %s: %s" % (obj, exc))
+        raise CommClosedError("in %s: %s" % (obj, exc)) from exc
+
+
+def _do_nothing():
+    pass
 
 
 class TCP(Comm):
     """
     An established communication based on an underlying Tornado IOStream.
     """
-    _iostream_allows_memoryview = tornado.version_info >= (4, 5)
-    # IOStream.read_into() currently proposed in
-    # https://github.com/tornadoweb/tornado/pull/2193
-    _iostream_has_read_into = hasattr(IOStream, "read_into")
 
     def __init__(self, stream, local_addr, peer_addr, deserialize=True):
+        Comm.__init__(self)
         self._local_addr = local_addr
         self._peer_addr = peer_addr
         self.stream = stream
         self.deserialize = deserialize
-        self._finalizer = finalize(self, self._get_finalizer())
+        self._finalizer = weakref.finalize(self, self._get_finalizer())
         self._finalizer.atexit = False
         self._extra = {}
 
         stream.set_nodelay(True)
         set_tcp_timeout(stream)
+        # set a close callback, to make `self.stream.closed()` more reliable.
+        # Background: if `stream` is unused (e.g. because it's in `ConnectionPool.available`),
+        # the underlying fd is not watched for changes. In this case, even if the
+        # connection is actively closed by the remote end, `self.closed()` would still return `False`.
+        # Registering a closed callback will make tornado register the underlying fd
+        # for changes, and this would be reflected in `self.closed()` even without reading/writing.
+        #
+        # Use a global method (instead of a lambda) to avoid creating a reference
+        # to the local scope.
+        stream.set_close_callback(_do_nothing)
         self._read_extra()
 
     def _read_extra(self):
@@ -170,99 +178,114 @@ class TCP(Comm):
     def peer_address(self):
         return self._peer_addr
 
-    @gen.coroutine
-    def read(self, deserializers=None):
+    async def read(self, deserializers=None):
         stream = self.stream
         if stream is None:
             raise CommClosedError
 
         try:
-            n_frames = yield stream.read_bytes(8)
-            n_frames = struct.unpack('Q', n_frames)[0]
-            lengths = yield stream.read_bytes(8 * n_frames)
-            lengths = struct.unpack('Q' * n_frames, lengths)
+            n_frames = await stream.read_bytes(8)
+            n_frames = struct.unpack("Q", n_frames)[0]
+            lengths = await stream.read_bytes(8 * n_frames)
+            lengths = struct.unpack("Q" * n_frames, lengths)
 
             frames = []
             for length in lengths:
-                if PY3 and self._iostream_has_read_into:
-                    frame = bytearray(length)
-                    if length:
-                        n = yield stream.read_into(frame)
-                        assert n == length, (n, length)
-                else:
-                    if length:
-                        frame = yield stream.read_bytes(length)
-                    else:
-                        frame = b''
+                frame = bytearray(length)
+                if length:
+                    n = await stream.read_into(frame)
+                    assert n == length, (n, length)
                 frames.append(frame)
         except StreamClosedError as e:
             self.stream = None
             if not shutting_down():
                 convert_stream_closed_error(self, e)
+        except Exception:
+            # Some OSError or a another "low-level" exception. We do not really know what
+            # was already read from the underlying socket, so it is not even safe to retry
+            # here using the same stream. The only safe thing to do is to abort.
+            # (See also GitHub #4133).
+            self.abort()
+            raise
         else:
             try:
-                msg = yield from_frames(frames,
-                                        deserialize=self.deserialize,
-                                        deserializers=deserializers)
+                msg = await from_frames(
+                    frames,
+                    deserialize=self.deserialize,
+                    deserializers=deserializers,
+                    allow_offload=self.allow_offload,
+                )
             except EOFError:
                 # Frames possibly garbled or truncated by communication error
                 self.abort()
                 raise CommClosedError("aborted stream on truncated data")
-            raise gen.Return(msg)
+            return msg
 
-    @gen.coroutine
-    def write(self, msg, serializers=None, on_error='message'):
+    async def write(self, msg, serializers=None, on_error="message"):
         stream = self.stream
         bytes_since_last_yield = 0
         if stream is None:
             raise CommClosedError
 
-        frames = yield to_frames(msg,
-                                 serializers=serializers,
-                                 on_error=on_error,
-                                 context={'sender': self._local_addr,
-                                          'recipient': self._peer_addr})
+        frames = await to_frames(
+            msg,
+            allow_offload=self.allow_offload,
+            serializers=serializers,
+            on_error=on_error,
+            context={
+                "sender": self.local_info,
+                "recipient": self.remote_info,
+                **self.handshake_options,
+            },
+        )
 
         try:
+            nframes = len(frames)
             lengths = [nbytes(frame) for frame in frames]
-            length_bytes = ([struct.pack('Q', len(frames))] +
-                            [struct.pack('Q', x) for x in lengths])
-            if PY3 and sum(lengths) < 2**17:  # 128kiB
-                b = b''.join(length_bytes + frames)  # small enough, send in one go
-                stream.write(b)
+            length_bytes = struct.pack(f"Q{nframes}Q", nframes, *lengths)
+            if sum(lengths) < 2 ** 17:  # 128kiB
+                # small enough, send in one go
+                stream.write(b"".join([length_bytes, *frames]))
             else:
-                stream.write(b''.join(length_bytes))  # avoid large memcpy, send in many
+                # avoid large memcpy, send in many
+                stream.write(length_bytes)
 
-                for frame in frames:
+                for frame, frame_bytes in zip(frames, lengths):
                     # Can't wait for the write() Future as it may be lost
                     # ("If write is called again before that Future has resolved,
                     #   the previous future will be orphaned and will never resolve")
-                    if not self._iostream_allows_memoryview:
-                        frame = ensure_bytes(frame)
                     future = stream.write(frame)
-                    bytes_since_last_yield += nbytes(frame)
+                    bytes_since_last_yield += frame_bytes
                     if bytes_since_last_yield > 32e6:
-                        yield future
+                        await future
                         bytes_since_last_yield = 0
         except StreamClosedError as e:
-            stream = None
-            convert_stream_closed_error(self, e)
-        except TypeError as e:
+            self.stream = None
+            if not shutting_down():
+                convert_stream_closed_error(self, e)
+        except Exception:
+            # Some OSError or a another "low-level" exception. We do not really know what
+            # was already written to the underlying socket, so it is not even safe to retry
+            # here using the same stream. The only safe thing to do is to abort.
+            # (See also GitHub #4133).
             if stream._write_buffer is None:
                 logger.info("tried to write message %s on closed stream", msg)
-            else:
-                raise
+            self.abort()
+            raise
 
-        raise gen.Return(sum(map(nbytes, frames)))
+        return sum(lengths)
 
     @gen.coroutine
     def close(self):
+        # We use gen.coroutine here rather than async def to avoid errors like
+        # Task was destroyed but it is pending!
+        # Triggered by distributed.deploy.tests.test_local::test_silent_startup
         stream, self.stream = self.stream, None
         if stream is not None and not stream.closed():
             try:
                 # Flush the stream's write buffer by waiting for a last write.
                 if stream.writing():
-                    yield stream.write(b'')
+                    yield stream.write(b"")
                 stream.socket.shutdown(socket.SHUT_RDWR)
             except EnvironmentError:
                 pass
@@ -293,45 +316,52 @@ class TLS(TCP):
         TCP._read_extra(self)
         sock = self.stream.socket
         if sock is not None:
-            self._extra.update(peercert=sock.getpeercert(),
-                               cipher=sock.cipher())
-            cipher, proto, bits = self._extra['cipher']
-            logger.debug("TLS connection with %r: protocol=%s, cipher=%s, bits=%d",
-                         self._peer_addr, proto, cipher, bits)
+            self._extra.update(peercert=sock.getpeercert(), cipher=sock.cipher())
+            cipher, proto, bits = self._extra["cipher"]
+            logger.debug(
+                "TLS connection with %r: protocol=%s, cipher=%s, bits=%d",
+                self._peer_addr,
+                proto,
+                cipher,
+                bits,
+            )
 
 
 def _expect_tls_context(connection_args):
-    ctx = connection_args.get('ssl_context')
+    ctx = connection_args.get("ssl_context")
     if not isinstance(ctx, ssl.SSLContext):
-        raise TypeError("TLS expects a `ssl_context` argument of type "
-                        "ssl.SSLContext (perhaps check your TLS configuration?)"
-                        "  Instead got %s" % str(ctx))
+        raise TypeError(
+            "TLS expects a `ssl_context` argument of type "
+            "ssl.SSLContext (perhaps check your TLS configuration?)"
+            "  Instead got %s" % str(ctx)
+        )
     return ctx
 
 
-class RequireEncryptionMixin(object):
-
+class RequireEncryptionMixin:
     def _check_encryption(self, address, connection_args):
-        if not self.encrypted and connection_args.get('require_encryption'):
+        if not self.encrypted and connection_args.get("require_encryption"):
             # XXX Should we have a dedicated SecurityError class?
-            raise RuntimeError("encryption required by Dask configuration, "
-                               "refusing communication from/to %r"
-                               % (self.prefix + address,))
+            raise RuntimeError(
+                "encryption required by Dask configuration, "
+                "refusing communication from/to %r" % (self.prefix + address,)
+            )
 
 
 class BaseTCPConnector(Connector, RequireEncryptionMixin):
+    _executor = ThreadPoolExecutor(2, thread_name_prefix="TCP-Executor")
+    _resolver = netutil.ExecutorResolver(close_executor=False, executor=_executor)
+    client = TCPClient(resolver=_resolver)
 
-    @gen.coroutine
-    def connect(self, address, deserialize=True, **connection_args):
+    async def connect(self, address, deserialize=True, **connection_args):
         self._check_encryption(address, connection_args)
         ip, port = parse_host_port(address)
         kwargs = self._get_connect_args(**connection_args)
 
-        client = TCPClient()
         try:
-            stream = yield client.connect(ip, port,
-                                          max_buffer_size=MAX_BUFFER_SIZE,
-                                          **kwargs)
+            stream = await self.client.connect(
+                ip, port, max_buffer_size=MAX_BUFFER_SIZE, **kwargs
+            )
             # Under certain circumstances tornado will have a closed connnection with an error and not raise
             # a StreamClosedError.
             #
@@ -342,16 +372,19 @@ class BaseTCPConnector(Connector, RequireEncryptionMixin):
         except StreamClosedError as e:
             # The socket connect() call failed
             convert_stream_closed_error(self, e)
+        except SSLError as err:
+            raise FatalCommClosedError() from err
 
         local_address = self.prefix + get_stream_address(stream)
-        raise gen.Return(self.comm_class(stream,
-                                         local_address,
-                                         self.prefix + address,
-                                         deserialize))
+        comm = self.comm_class(
+            stream, local_address, self.prefix + address, deserialize
+        )
+
+        return comm
 
 
 class TCPConnector(BaseTCPConnector):
-    prefix = 'tcp://'
+    prefix = "tcp://"
     comm_class = TCP
     encrypted = False
 
@@ -360,40 +393,47 @@ class TCPConnector(BaseTCPConnector):
 
 
 class TLSConnector(BaseTCPConnector):
-    prefix = 'tls://'
+    prefix = "tls://"
     comm_class = TLS
     encrypted = True
 
     def _get_connect_args(self, **connection_args):
         ctx = _expect_tls_context(connection_args)
-        return {'ssl_options': ctx}
+        return {"ssl_options": ctx}
 
 
 class BaseTCPListener(Listener, RequireEncryptionMixin):
-
-    def __init__(self, address, comm_handler, deserialize=True,
-                 default_port=0, **connection_args):
+    def __init__(
+        self,
+        address,
+        comm_handler,
+        deserialize=True,
+        allow_offload=True,
+        default_port=0,
+        **connection_args,
+    ):
         self._check_encryption(address, connection_args)
         self.ip, self.port = parse_host_port(address, default_port)
         self.comm_handler = comm_handler
         self.deserialize = deserialize
+        self.allow_offload = allow_offload
         self.server_args = self._get_server_args(**connection_args)
         self.tcp_server = None
         self.bound_address = None
 
-    def start(self):
-        self.tcp_server = TCPServer(max_buffer_size=MAX_BUFFER_SIZE,
-                                    **self.server_args)
+    async def start(self):
+        self.tcp_server = TCPServer(max_buffer_size=MAX_BUFFER_SIZE, **self.server_args)
         self.tcp_server.handle_stream = self._handle_stream
-        backlog = int(dask.config.get('distributed.comm.socket-backlog'))
+        backlog = int(dask.config.get("distributed.comm.socket-backlog"))
         for i in range(5):
             try:
                 # When shuffling data between workers, there can
                 # really be O(cluster size) connection requests
                 # on a single worker socket, make sure the backlog
                 # is large enough not to lose any.
-                sockets = netutil.bind_sockets(self.port, address=self.ip,
-                                               backlog=backlog)
+                sockets = netutil.bind_sockets(
+                    self.port, address=self.ip, backlog=backlog
+                )
             except EnvironmentError as e:
                 # EADDRINUSE can happen sporadically when trying to bind
                 # to an ephemeral port
@@ -405,6 +445,7 @@ class BaseTCPListener(Listener, RequireEncryptionMixin):
                 break
         else:
             raise exc
+        self.get_host_port()  # trigger assignment to self.bound_address
 
     def stop(self):
         tcp_server, self.tcp_server = self.tcp_server, None
@@ -415,18 +456,24 @@ class BaseTCPListener(Listener, RequireEncryptionMixin):
         if self.tcp_server is None:
             raise ValueError("invalid operation on non-started TCPListener")
 
-    @gen.coroutine
-    def _handle_stream(self, stream, address):
+    async def _handle_stream(self, stream, address):
         address = self.prefix + unparse_host_port(*address[:2])
-        stream = yield self._prepare_stream(stream, address)
+        stream = await self._prepare_stream(stream, address)
         if stream is None:
             # Preparation failed
             return
-        logger.debug("Incoming connection from %r to %r",
-                     address, self.contact_address)
+        logger.debug("Incoming connection from %r to %r", address, self.contact_address)
         local_address = self.prefix + get_stream_address(stream)
         comm = self.comm_class(stream, local_address, address, self.deserialize)
-        self.comm_handler(comm)
+        comm.allow_offload = self.allow_offload
+
+        try:
+            await self.on_connection(comm)
+        except CommClosedError:
+            logger.info("Connection closed before handshake completed")
+            return
+
+        await self.comm_handler(comm)
 
     def get_host_port(self):
         """
@@ -457,38 +504,39 @@ class BaseTCPListener(Listener, RequireEncryptionMixin):
 
 
 class TCPListener(BaseTCPListener):
-    prefix = 'tcp://'
+    prefix = "tcp://"
     comm_class = TCP
     encrypted = False
 
     def _get_server_args(self, **connection_args):
         return {}
 
-    @gen.coroutine
-    def _prepare_stream(self, stream, address):
-        raise gen.Return(stream)
+    async def _prepare_stream(self, stream, address):
+        return stream
 
 
 class TLSListener(BaseTCPListener):
-    prefix = 'tls://'
+    prefix = "tls://"
     comm_class = TLS
     encrypted = True
 
     def _get_server_args(self, **connection_args):
         ctx = _expect_tls_context(connection_args)
-        return {'ssl_options': ctx}
+        return {"ssl_options": ctx}
 
-    @gen.coroutine
-    def _prepare_stream(self, stream, address):
+    async def _prepare_stream(self, stream, address):
         try:
-            yield stream.wait_for_handshake()
+            await stream.wait_for_handshake()
         except EnvironmentError as e:
             # The handshake went wrong, log and ignore
-            logger.warning("Listener on %r: TLS handshake failed with remote %r: %s",
-                           self.listen_address, address,
-                           getattr(e, "real_error", None) or e)
+            logger.warning(
+                "Listener on %r: TLS handshake failed with remote %r: %s",
+                self.listen_address,
+                address,
+                getattr(e, "real_error", None) or e,
+            )
         else:
-            raise gen.Return(stream)
+            return stream
 
 
 class BaseTCPBackend(Backend):
@@ -516,7 +564,7 @@ class BaseTCPBackend(Backend):
     def get_local_address_for(self, loc):
         host, port = parse_host_port(loc)
         host = ensure_ip(host)
-        if ':' in host:
+        if ":" in host:
             local_host = get_ipv6(host)
         else:
             local_host = get_ip(host)
@@ -533,5 +581,5 @@ class TLSBackend(BaseTCPBackend):
     _listener_class = TLSListener
 
 
-backends['tcp'] = TCPBackend()
-backends['tls'] = TLSBackend()
+backends["tcp"] = TCPBackend()
+backends["tls"] = TLSBackend()
