@@ -1,9 +1,8 @@
-
 import numpy as np
 from numpy.testing import assert_allclose
 import pytest
 
-keras = pytest.importorskip('keras')
+keras = pytest.importorskip("keras")
 
 from distributed.protocol import serialize, deserialize, dumps, loads, to_serialize
 
@@ -12,7 +11,7 @@ def test_serialize_deserialize_model():
     model = keras.models.Sequential()
     model.add(keras.layers.Dense(5, input_dim=3))
     model.add(keras.layers.Dense(2))
-    model.compile(optimizer='sgd', loss='mse')
+    model.compile(optimizer="sgd", loss="mse")
     x = np.random.random((1, 3))
     y = np.random.random((1, 2))
     model.train_on_batch(x, y)
@@ -20,7 +19,7 @@ def test_serialize_deserialize_model():
     loaded = deserialize(*serialize(model))
     assert_allclose(loaded.predict(x), model.predict(x))
 
-    data = {'model': to_serialize(model)}
+    data = {"model": to_serialize(model)}
     frames = dumps(data)
     result = loads(frames)
-    assert_allclose(result['model'].predict(x), model.predict(x))
+    assert_allclose(result["model"].predict(x), model.predict(x))
