@@ -1,9 +1,8 @@
-
 import numpy as np
 from numpy.testing import assert_allclose
 import pytest
 
-sparse = pytest.importorskip('sparse')
+sparse = pytest.importorskip("sparse")
 
 from distributed.protocol import deserialize, serialize
 
@@ -14,7 +13,7 @@ def test_serialize_deserialize_sparse():
 
     y = sparse.COO(x)
     header, frames = serialize(y)
-    assert 'sparse' in header['type']
+    assert "sparse" in header["type"]
     z = deserialize(*serialize(y))
 
     assert_allclose(y.data, z.data)
