@@ -1,13 +1,13 @@
 import asyncio
 from asyncio import TimeoutError
 import atexit
+import click
 from collections import deque, OrderedDict, UserDict
-from concurrent.futures import ThreadPoolExecutor
-from contextlib import contextmanager
+from concurrent.futures import ThreadPoolExecutor, CancelledError  # noqa: F401
+from contextlib import contextmanager, suppress
 import functools
 from hashlib import md5
 import html
-import inspect
 import json
 import logging
 import multiprocessing
@@ -46,8 +46,7 @@ from dask.utils import (  # noqa
     parse_timedelta,
 )
 
-import toolz
-import tornado
+import tlz as toolz
 from tornado import gen
 from tornado.ioloop import IOLoop
 
@@ -138,13 +137,15 @@ def _get_ip(host, port, family):
         ip = sock.getsockname()[0]
         return ip
     except EnvironmentError as e:
-        # XXX Should first try getaddrinfo() on socket.gethostname() and getfqdn()
         warnings.warn(
             "Couldn't detect a suitable IP address for "
             "reaching %r, defaulting to hostname: %s" % (host, e),
             RuntimeWarning,
         )
-        return socket.gethostname()
+        addr_info = socket.getaddrinfo(
+            socket.gethostname(), port, family, socket.SOCK_DGRAM, socket.IPPROTO_UDP
+        )[0]
+        return addr_info[4][0]
     finally:
         sock.close()
 
@@ -191,17 +192,10 @@ def get_ip_interface(ifname):
     raise ValueError("interface %r doesn't have an IPv4 address" % (ifname,))
 
 
-@contextmanager
-def ignoring(*exceptions):
-    try:
-        yield
-    except exceptions as e:
-        pass
-
-
+# FIXME: this breaks if changed to async def...
 @gen.coroutine
 def ignore_exceptions(coroutines, *exceptions):
-    """ Process list of coroutines, ignoring certain exceptions
+    """Process list of coroutines, ignoring certain exceptions
 
     >>> coroutines = [cor(...) for ...]  # doctest: +SKIP
     >>> x = yield ignore_exceptions(coroutines, TypeError)  # doctest: +SKIP
@@ -209,14 +203,14 @@ def ignore_exceptions(coroutines, *exceptions):
     wait_iterator = gen.WaitIterator(*coroutines)
     results = []
     while not wait_iterator.done():
-        with ignoring(*exceptions):
+        with suppress(*exceptions):
             result = yield wait_iterator.next()
             results.append(result)
     raise gen.Return(results)
 
 
 async def All(args, quiet_exceptions=()):
-    """ Wait on many tasks at the same time
+    """Wait on many tasks at the same time
 
     Err once any of the tasks err.
 
@@ -237,7 +231,7 @@ async def All(args, quiet_exceptions=()):
 
             @gen.coroutine
             def quiet():
-                """ Watch unfinished tasks
+                """Watch unfinished tasks
 
                 Otherwise if they err they get logged in a way that is hard to
                 control.  They need some other task to watch them so that they
@@ -256,7 +250,7 @@ async def All(args, quiet_exceptions=()):
 
 
 async def Any(args, quiet_exceptions=()):
-    """ Wait on many tasks at the same time and return when any is finished
+    """Wait on many tasks at the same time and return when any is finished
 
     Err once any of the tasks err.
 
@@ -275,7 +269,7 @@ async def Any(args, quiet_exceptions=()):
 
             @gen.coroutine
             def quiet():
-                """ Watch unfinished tasks
+                """Watch unfinished tasks
 
                 Otherwise if they err they get logged in a way that is hard to
                 control.  They need some other task to watch them so that they
@@ -299,6 +293,7 @@ def sync(loop, func, *args, callback_timeout=None, **kwargs):
     """
     Run coroutine in loop running in separate thread.
     """
+    callback_timeout = parse_timedelta(callback_timeout, "s")
     # Tornado's PollIOLoop doesn't raise when using closed, do it ourselves
     if PollIOLoop and (
         (isinstance(loop, PollIOLoop) and getattr(loop, "_closing", False))
@@ -347,7 +342,7 @@ def sync(loop, func, *args, callback_timeout=None, **kwargs):
         return result[0]
 
 
-class LoopRunner(object):
+class LoopRunner:
     """
     A helper to start and stop an IO loop in a controlled way.
     Several loop runners can associate safely to the same IO loop.
@@ -377,10 +372,8 @@ class LoopRunner(object):
                 # We're expecting the loop to run in another thread,
                 # avoid re-using this thread's assigned loop
                 self._loop = IOLoop()
-            self._should_close_loop = True
         else:
             self._loop = loop
-            self._should_close_loop = False
         self._asynchronous = asynchronous
         self._loop_thread = None
         self._started = False
@@ -479,7 +472,7 @@ class LoopRunner(object):
             try:
                 self._loop.add_callback(self._loop.stop)
                 self._loop_thread.join(timeout=timeout)
-                with ignoring(KeyError):  # IOLoop can be missing
+                with suppress(KeyError):  # IOLoop can be missing
                     self._loop.close()
             finally:
                 self._loop_thread = None
@@ -550,7 +543,7 @@ def clear_queue(q):
 
 
 def is_kernel():
-    """ Determine if we're running within an IPython kernel
+    """Determine if we're running within an IPython kernel
 
     >>> is_kernel()
     False
@@ -647,11 +640,11 @@ def key_split_group(x):
         elif x[0] == "<":
             return x.strip("<>").split()[0].split(".")[-1]
         else:
-            return ""
+            return key_split(x)
     elif typ is bytes:
         return key_split_group(x.decode())
     else:
-        return ""
+        return key_split(x)
 
 
 @contextmanager
@@ -693,7 +686,7 @@ def silence_logging(level, root="distributed"):
 
 @toolz.memoize
 def ensure_ip(hostname):
-    """ Ensure that address is an IP address
+    """Ensure that address is an IP address
 
     Examples
     --------
@@ -746,29 +739,8 @@ def truncate_exception(e, n=10000):
         return e
 
 
-def tokey(o):
-    """ Convert an object to a string.
-
-    Examples
-    --------
-
-    >>> tokey(b'x')
-    b'x'
-    >>> tokey('x')
-    'x'
-    >>> tokey(1)
-    '1'
-    """
-    typ = type(o)
-    if typ is str or typ is bytes:
-        return o
-    else:
-        return str(o)
-
-
 def validate_key(k):
-    """Validate a key as received on a stream.
-    """
+    """Validate a key as received on a stream."""
     typ = type(k)
     if typ is not str and typ is not bytes:
         raise TypeError("Unexpected key type %s (value: %r)" % (typ, k))
@@ -785,27 +757,8 @@ def _maybe_complex(task):
     )
 
 
-def convert(task, dsk, extra_values):
-    if type(task) is list:
-        return [convert(v, dsk, extra_values) for v in task]
-    if type(task) is dict:
-        return {k: convert(v, dsk, extra_values) for k, v in task.items()}
-    if istask(task):
-        return (task[0],) + tuple(convert(x, dsk, extra_values) for x in task[1:])
-    try:
-        if task in dsk or task in extra_values:
-            return tokey(task)
-    except TypeError:
-        pass
-    return task
-
-
-def str_graph(dsk, extra_values=()):
-    return {tokey(k): convert(v, dsk, extra_values) for k, v in dsk.items()}
-
-
 def seek_delimiter(file, delimiter, blocksize):
-    """ Seek current file to next byte after a delimiter bytestring
+    """Seek current file to next byte after a delimiter bytestring
 
     This seeks the file to the next byte following the delimiter.  It does
     not return anything.  Use ``file.tell()`` to see location afterwards.
@@ -838,7 +791,7 @@ def seek_delimiter(file, delimiter, blocksize):
 
 
 def read_block(f, offset, length, delimiter=None):
-    """ Read a block of bytes from a file
+    """Read a block of bytes from a file
 
     Parameters
     ----------
@@ -937,7 +890,9 @@ def ensure_bytes(s):
     >>> ensure_bytes(b'123')
     b'123'
     """
-    if hasattr(s, "encode"):
+    if isinstance(s, bytes):
+        return s
+    elif hasattr(s, "encode"):
         return s.encode()
     else:
         try:
@@ -1004,7 +959,7 @@ shutting_down.__doc__ = """
 
 
 def open_port(host=""):
-    """ Return a probably-open port
+    """Return a probably-open port
 
     There is a chance that this port will be taken by the operating system soon
     after returning from this function.
@@ -1031,15 +986,12 @@ def import_file(path):
         names_to_import.append(name)
     if ext == ".py":  # Ensure that no pyc file will be reused
         cache_file = cache_from_source(path)
-        with ignoring(OSError):
+        with suppress(OSError):
             os.remove(cache_file)
     if ext in (".egg", ".zip", ".pyz"):
         if path not in sys.path:
             sys.path.insert(0, path)
-        if sys.version_info >= (3, 6):
-            names = (mod_info.name for mod_info in pkgutil.iter_modules([path]))
-        else:
-            names = (mod_info[1] for mod_info in pkgutil.iter_modules([path]))
+        names = (mod_info.name for mod_info in pkgutil.iter_modules([path]))
         names_to_import.extend(names)
 
     loaded = []
@@ -1059,7 +1011,7 @@ def import_file(path):
     return loaded
 
 
-class itemgetter(object):
+class itemgetter:
     """A picklable itemgetter.
 
     Examples
@@ -1114,15 +1066,17 @@ def nbytes(frame, _bytes_like=(bytes, bytearray)):
             return len(frame)
 
 
-def PeriodicCallback(callback, callback_time, io_loop=None):
+def is_writeable(frame):
     """
-    Wrapper around tornado.IOLoop.PeriodicCallback, for compatibility
-    with removal of the `io_loop` parameter in Tornado 5.0.
+    Check whether frame is writeable
+
+    Will return ``True`` if writeable, ``False`` if readonly, and
+    ``None`` if undetermined.
     """
-    if tornado.version_info >= (5,):
-        return tornado.ioloop.PeriodicCallback(callback, callback_time)
-    else:
-        return tornado.ioloop.PeriodicCallback(callback, callback_time, io_loop)
+    try:
+        return not memoryview(frame).readonly
+    except TypeError:
+        return None
 
 
 @contextmanager
@@ -1156,7 +1110,7 @@ class DequeHandler(logging.Handler):
 
     def __init__(self, *args, n=10000, **kwargs):
         self.deque = deque(maxlen=n)
-        super(DequeHandler, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self._instances.add(self)
 
     def emit(self, record):
@@ -1178,7 +1132,7 @@ class DequeHandler(logging.Handler):
 
 
 def reset_logger_locks():
-    """ Python 2's logger's locks don't survive a fork event
+    """Python 2's logger's locks don't survive a fork event
 
     https://github.com/dask/distributed/issues/1491
     """
@@ -1187,33 +1141,47 @@ def reset_logger_locks():
             handler.createLock()
 
 
-if tornado.version_info[0] >= 5:
+is_server_extension = False
 
-    is_server_extension = False
+if "notebook" in sys.modules:
+    import traitlets
+    from notebook.notebookapp import NotebookApp
 
-    if "notebook" in sys.modules:
-        import traitlets
-        from notebook.notebookapp import NotebookApp
+    is_server_extension = traitlets.config.Application.initialized() and isinstance(
+        traitlets.config.Application.instance(), NotebookApp
+    )
 
-        is_server_extension = traitlets.config.Application.initialized() and isinstance(
-            traitlets.config.Application.instance(), NotebookApp
-        )
+if not is_server_extension:
+    is_kernel_and_no_running_loop = False
 
-    if not is_server_extension:
-        is_kernel_and_no_running_loop = False
+    if is_kernel():
+        try:
+            get_running_loop()
+        except RuntimeError:
+            is_kernel_and_no_running_loop = True
 
-        if is_kernel():
-            try:
-                get_running_loop()
-            except RuntimeError:
-                is_kernel_and_no_running_loop = True
+    if not is_kernel_and_no_running_loop:
 
-        if not is_kernel_and_no_running_loop:
-            import tornado.platform.asyncio
+        # TODO: Use tornado's AnyThreadEventLoopPolicy, instead of class below,
+        # once tornado > 6.0.3 is available.
+        if WINDOWS and hasattr(asyncio, "WindowsSelectorEventLoopPolicy"):
+            # WindowsProactorEventLoopPolicy is not compatible with tornado 6
+            # fallback to the pre-3.8 default of Selector
+            # https://github.com/tornadoweb/tornado/issues/2608
+            BaseEventLoopPolicy = asyncio.WindowsSelectorEventLoopPolicy
+        else:
+            BaseEventLoopPolicy = asyncio.DefaultEventLoopPolicy
 
-            asyncio.set_event_loop_policy(
-                tornado.platform.asyncio.AnyThreadEventLoopPolicy()
-            )
+        class AnyThreadEventLoopPolicy(BaseEventLoopPolicy):
+            def get_event_loop(self):
+                try:
+                    return super().get_event_loop()
+                except (RuntimeError, AssertionError):
+                    loop = self.new_event_loop()
+                    self.set_event_loop(loop)
+                    return loop
+
+        asyncio.set_event_loop_policy(AnyThreadEventLoopPolicy())
 
 
 @functools.lru_cache(1000)
@@ -1221,6 +1189,32 @@ def has_keyword(func, keyword):
     return keyword in inspect.signature(func).parameters
 
 
+@functools.lru_cache(1000)
+def command_has_keyword(cmd, k):
+    if cmd is not None:
+        if isinstance(cmd, str):
+            try:
+                from importlib import import_module
+
+                cmd = import_module(cmd)
+            except ImportError:
+                raise ImportError("Module for command %s is not available" % cmd)
+
+        if isinstance(getattr(cmd, "main"), click.core.Command):
+            cmd = cmd.main
+        if isinstance(cmd, click.core.Command):
+            cmd_params = set(
+                [
+                    p.human_readable_name
+                    for p in cmd.params
+                    if isinstance(p, click.core.Option)
+                ]
+            )
+            return k in cmd_params
+
+    return False
+
+
 # from bokeh.palettes import viridis
 # palette = viridis(18)
 palette = [
@@ -1253,11 +1247,7 @@ def color_of(x, palette=palette):
 
 
 def iscoroutinefunction(f):
-    if gen.is_coroutine_function(f):
-        return True
-    if sys.version_info >= (3, 5) and inspect.iscoroutinefunction(f):
-        return True
-    return False
+    return inspect.iscoroutinefunction(f) or gen.is_coroutine_function(f)
 
 
 @contextmanager
@@ -1270,7 +1260,7 @@ def warn_on_duration(duration, msg):
 
 
 def typename(typ):
-    """ Return name of type
+    """Return name of type
 
     Examples
     --------
@@ -1295,6 +1285,64 @@ def format_dashboard_link(host, port):
     )
 
 
+def parse_ports(port):
+    """Parse input port information into list of ports
+
+    Parameters
+    ----------
+    port : int, str, None
+        Input port or ports. Can be an integer like 8787, a string for a
+        single port like "8787", a string for a sequential range of ports like
+        "8000:8200", or None.
+
+    Returns
+    -------
+    ports : list
+        List of ports
+
+    Examples
+    --------
+    A single port can be specified using an integer:
+
+    >>> parse_ports(8787)
+    >>> [8787]
+
+    or a string:
+
+    >>> parse_ports("8787")
+    >>> [8787]
+
+    A sequential range of ports can be specified by a string which indicates
+    the first and last ports which should be included in the sequence of ports:
+
+    >>> parse_ports("8787:8790")
+    >>> [8787, 8788, 8789, 8790]
+
+    An input of ``None`` is also valid and can be used to indicate that no port
+    has been specified:
+
+    >>> parse_ports(None)
+    >>> [None]
+
+    """
+    if isinstance(port, str) and ":" not in port:
+        port = int(port)
+
+    if isinstance(port, (int, type(None))):
+        ports = [port]
+    else:
+        port_start, port_stop = map(int, port.split(":"))
+        if port_stop <= port_start:
+            raise ValueError(
+                "When specifying a range of ports like port_start:port_stop, "
+                "port_stop must be greater than port_start, but got "
+                f"port_start={port_start} and port_stop={port_stop}"
+            )
+        ports = list(range(port_start, port_stop + 1))
+
+    return ports
+
+
 def is_coroutine_function(f):
     return asyncio.iscoroutinefunction(f) or gen.is_coroutine_function(f)
 
@@ -1322,8 +1370,8 @@ class Logs(dict):
         return "\n".join(summaries)
 
 
-def cli_keywords(d: dict, cls=None):
-    """ Convert a kwargs dictionary into a list of CLI keywords
+def cli_keywords(d: dict, cls=None, cmd=None):
+    """Convert a kwargs dictionary into a list of CLI keywords
 
     Parameters
     ----------
@@ -1331,6 +1379,12 @@ def cli_keywords(d: dict, cls=None):
         The keywords to convert
     cls: callable
         The callable that consumes these terms to check them for validity
+    cmd: string or object
+        A string with the name of a module, or the module containing a
+        click-generated command with a "main" function, or the function itself.
+        It may be used to parse a module's custom arguments (i.e., arguments that
+        are not part of Worker class), such as nprocs from dask-worker CLI or
+        enable_nvlink from dask-cuda-worker CLI.
 
     Examples
     --------
@@ -1343,12 +1397,22 @@ def cli_keywords(d: dict, cls=None):
     ...
     ValueError: Class distributed.worker.Worker does not support keyword x
     """
-    if cls:
+    if cls or cmd:
         for k in d:
-            if not has_keyword(cls, k):
-                raise ValueError(
-                    "Class %s does not support keyword %s" % (typename(cls), k)
-                )
+            if not has_keyword(cls, k) and not command_has_keyword(cmd, k):
+                if cls and cmd:
+                    raise ValueError(
+                        "Neither class %s or module %s support keyword %s"
+                        % (typename(cls), typename(cmd), k)
+                    )
+                elif cls:
+                    raise ValueError(
+                        "Class %s does not support keyword %s" % (typename(cls), k)
+                    )
+                else:
+                    raise ValueError(
+                        "Module %s does not support keyword %s" % (typename(cmd), k)
+                    )
 
     def convert_value(v):
         out = str(v)
@@ -1365,18 +1429,12 @@ def is_valid_xml(text):
     return xml.etree.ElementTree.fromstring(text) is not None
 
 
-try:
-    _offload_executor = ThreadPoolExecutor(
-        max_workers=1, thread_name_prefix="Dask-Offload"
-    )
-except TypeError:
-    _offload_executor = ThreadPoolExecutor(max_workers=1)
-
+_offload_executor = ThreadPoolExecutor(max_workers=1, thread_name_prefix="Dask-Offload")
 weakref.finalize(_offload_executor, _offload_executor.shutdown)
 
 
 def import_term(name: str):
-    """ Return the fully qualified term
+    """Return the fully qualified term
 
     Examples
     --------
@@ -1398,7 +1456,7 @@ async def offload(fn, *args, **kwargs):
 
 
 def serialize_for_cli(data):
-    """ Serialize data into a string that can be passthrough cli
+    """Serialize data into a string that can be passthrough cli
 
     Parameters
     ----------
@@ -1413,7 +1471,7 @@ def serialize_for_cli(data):
 
 
 def deserialize_for_cli(data):
-    """ De-serialize data into the original object
+    """De-serialize data into the original object
 
     Parameters
     ----------
@@ -1445,8 +1503,7 @@ empty_context = EmptyContext()
 
 
 class LRU(UserDict):
-    """ Limited size mapping, evicting the least recently looked-up key when full
-    """
+    """Limited size mapping, evicting the least recently looked-up key when full"""
 
     def __init__(self, maxsize):
         super().__init__()
@@ -1462,3 +1519,45 @@ class LRU(UserDict):
         if len(self) >= self.maxsize:
             self.data.popitem(last=False)
         super().__setitem__(key, value)
+
+
+def clean_dashboard_address(addr, default_listen_ip=""):
+    """
+
+    Examples
+    --------
+    >>> clean_dashboard_address(8787)
+    {'address': '', 'port': 8787}
+    >>> clean_dashboard_address(":8787")
+    {'address': '', 'port': 8787}
+    >>> clean_dashboard_address("8787")
+    {'address': '', 'port': 8787}
+    >>> clean_dashboard_address("8787")
+    {'address': '', 'port': 8787}
+    >>> clean_dashboard_address("foo:8787")
+    {'address': 'foo', 'port': 8787}
+    """
+
+    if default_listen_ip == "0.0.0.0":
+        default_listen_ip = ""  # for IPV6
+
+    try:
+        addr = int(addr)
+    except (TypeError, ValueError):
+        pass
+
+    if isinstance(addr, str):
+        addr = addr.split(":")
+
+    if isinstance(addr, (tuple, list)):
+        if len(addr) == 2:
+            host, port = (addr[0], int(addr[1]))
+        elif len(addr) == 1:
+            [host], port = addr, 0
+        else:
+            raise ValueError(addr)
+    elif isinstance(addr, int):
+        host = default_listen_ip
+        port = addr
+
+    return {"address": host, "port": port}
